
#include "globals.h"

void parseLIST( 
#if NeedFunctionPrototypes
	char*, struct FileManager*, int
#endif
);


int ftpConnect( f )
  struct FileManager* f;
{
  static int            	firstTime = 1;
  static struct servent *       sp;

  struct sockaddr_in 	tcp_srv_addr;
  char			buff[100];

  if( firstTime )
  {
    if( (sp = getservbyname("ftp","tcp")) == NULL )
    {
      popupErrorDialog( "***ERROR: getservbyname() failed..." );
      return -1;
    }
    firstTime = 0;
  }

  memset( (void *) &tcp_srv_addr, 0, sizeof(tcp_srv_addr) );
  tcp_srv_addr.sin_family = AF_INET;
  tcp_srv_addr.sin_port = sp->s_port;

  if( (tcp_srv_addr.sin_addr.s_addr = inet_addr(f->host)) == INADDR_NONE )
  {
    struct hostent *hp;
    if( (hp = gethostbyname(f->host)) == NULL )
    {
      sprintf( buff, "***ERROR: gethostbyname() failed for host = %s", f->host );
      popupErrorDialog( buff );
      return -1;
    }
    memcpy( (void *) &tcp_srv_addr.sin_addr, (void *) hp->h_addr, hp->h_length );
  }

  if( (f->cmdSocketId = socket(AF_INET,SOCK_STREAM,0)) < 0 )
  {
    sprintf( buff, "***ERROR: socket() failed for host = %s", f->host );
    popupErrorDialog( buff );
    return -1; 
  } 

  if( connect( f->cmdSocketId, (struct sockaddr *) &tcp_srv_addr,
	sizeof(tcp_srv_addr) ) < 0 )
  {
    sprintf( buff, "***ERROR: connect() failed for host = %s", f->host );
    popupErrorDialog( buff );
    close( f->cmdSocketId );
    return -1;
  }

  f->cmdId = XtAppAddInput( XtWidgetToApplicationContext(f->parent), 
	f->cmdSocketId, (XtPointer)XtInputReadMask, readFtpMessages, 
	(XtPointer)f );

  return 0;
}

void ftpSendCommand( f, cmd, index, recvDir )
  struct FileManager* f;
  char* cmd, *recvDir;
  int index;
{
  char buffer[1056];

  printf( "ftpSendCommand() -> cmd = %s\n", cmd );
  fflush( stdout );

  if( !strncmp( cmd, "LIST", 4 ) || !strncmp( cmd, "STOR", 4 ) || 
      !strncmp( cmd, "RETR", 4 ) )
  {
    char buf[64];
    XtInputId dataId;
    struct sockaddr_in a;
    struct sockaddr_in c;
    int alen = sizeof(a);
    int clen = sizeof(c);
    struct FtpDataConn *ft;

    a.sin_family = AF_INET;
    a.sin_addr.s_addr = (u_int)INADDR_ANY;
    a.sin_port = 0;
    memset( (void *)(a.sin_zero), 0, sizeof(a.sin_zero) );

    ft = appendFtpDataConn( &f->ftpDataConn );
    ft->index = index;
    ft->fd = (FILE *)NULL;
    strcpy( ft->recvDir, recvDir );

    ft->results = (char *)XtMalloc( 1 );
    strcpy( ft->results, "" );
    ft->cmd = (char *)XtMalloc( strlen(cmd)+1 );
    strcpy( ft->cmd, cmd );

    ft->dataSocketId = socket( AF_INET, SOCK_STREAM, 0 );

    if( bind( ft->dataSocketId, (struct sockaddr *)&a, sizeof(a) ) < 0 )
    {
      sprintf( buffer, "***ERROR: bind() failed for host = %s", f->host );
      popupErrorDialog( buffer );
      return;
    }

    if( listen( ft->dataSocketId, 1 ) < 0 )
    {
      sprintf( buffer, "***ERROR: listen() failed for host = %s", f->host );
      popupErrorDialog( buffer );
      return;
    }

    dataId = XtAppAddInput( XtWidgetToApplicationContext(f->parent),
        ft->dataSocketId, (XtPointer)XtInputReadMask, establishFtpDataConn, 
	(XtPointer)f );

    if( getsockname( ft->dataSocketId, (struct sockaddr *)&a, &alen ) < 0 )
    {
      sprintf( buffer, "***ERROR: getsockname() failed for host = %s", f->host );
      popupErrorDialog( buffer );
      return;
    }

    if( getsockname( f->cmdSocketId, (struct sockaddr *)&c, &clen ) < 0 )
    {
      sprintf( buffer, "***ERROR: getsockname() failed for host = %s", f->host );
      popupErrorDialog( buffer );
      return;
    }

    sprintf(buf, "PORT %u,%u,%u,%u,%u,%u",
                 (htonl((u_int)(c.sin_addr.s_addr)) >> (3 * 8)) & 0xFF,
                 (htonl((u_int)(c.sin_addr.s_addr)) >> (2 * 8)) & 0xFF,
                 (htonl((u_int)(c.sin_addr.s_addr)) >> (1 * 8)) & 0xFF,
                 (htonl((u_int)(c.sin_addr.s_addr)) >> (0 * 8)) & 0xFF,
                 (htons((u_int)(a.sin_port)) >> (1 * 8)) & 0xFF,
                 (htons((u_int)(a.sin_port)) >> (0 * 8)) & 0xFF
                );

    write( f->cmdSocketId, buf, strlen(buf) );
    write( f->cmdSocketId, "\r\n", 2 );
  }

  write( f->cmdSocketId, cmd, strlen(cmd) );
  write( f->cmdSocketId, "\r\n", 2 );
}

void establishFtpDataConn( xtp, source, id )
  XtPointer xtp;
  int* source;
  XtInputId* id;
{
  struct FileManager *f = (struct FileManager *)xtp;
  struct sockaddr_in a;
  int i, len = sizeof(a), n = numFtpDataConn( f->ftpDataConn );
  struct FtpDataConn *ft;
  char buffer[1056];

  XtRemoveInput( *id );

  for( i = 0; i < n; i++ )
  {
    ft = getIthFtpDataConn( f->ftpDataConn, i );
    if( *source == ft->dataSocketId )  break;
  }

  if( (ft->acceptSocketId = accept( ft->dataSocketId, (struct sockaddr *)&a, 
	&len )) < 0 )
  {
    sprintf( buffer, "***ERROR: accept() failed for host = %s", f->host );
    popupErrorDialog( buffer );
    close( ft->dataSocketId );
    removeIthFtpDataConn( &f->ftpDataConn, i );
    return;
  }

/*
  if( !strncmp( ft->cmd, "RETR", 4 ) )
  {
    printf( "Opening %s\n", ft->recvDir );
    if( (ft->fd = fopen( ft->recvDir, "w" )) == (FILE *)NULL )
    {
      ft->fd = (FILE *)NULL;
      write( f->cmdSocketId, "ABOR\r\n", 6 );
      sprintf( buffer, "***ERROR: cannot open %s on local machine", ft->recvDir );
      popupErrorDialog( buffer );
      close( ft->dataSocketId );
      removeIthFtpDataConn( &f->ftpDataConn, i );
      return;
    }
  }
*/

  ft->acceptId = XtAppAddInput( XtWidgetToApplicationContext(f->parent),
	ft->acceptSocketId, (XtPointer)XtInputReadMask, readFtpData,
	(XtPointer)f );

  close( ft->dataSocketId );
  ft->dataSocketId = 0;
}

void readFtpData( xtp, source, id )
  XtPointer xtp;
  int* source;
  XtInputId* id;
{
  struct FileManager *f = (struct FileManager *)xtp;
  char buffer[1056];
  int n = numFtpDataConn( f->ftpDataConn ), nb = sizeof( buffer ), i;
  struct FtpDataConn *ft;

  memset( (void *)buffer, '\0', nb );

  for( i = 0; i < n; i++ )
  {
    ft = getIthFtpDataConn( f->ftpDataConn, i );
    if( *source == ft->acceptSocketId )  break;
  }

  if( !strncmp( ft->cmd, "RETR", 4 ) )
  {
    if( f->transferType == TRANSFER_BINARY )
    {
      if( (ft->fd = fopen( ft->recvDir, "a" )) == (FILE *)NULL )
      {
        write( f->cmdSocketId, "ABOR\r\n", 6 );
        sprintf( buffer, "***ERROR: file descriptor for %s is NULL", ft->recvDir );
        popupErrorDialog( buffer );
        close( ft->dataSocketId );
        removeIthFtpDataConn( &f->ftpDataConn, i );
        return;
      }
      else
      {
        FILE *fd;
	if( (fd = fdopen(*source,"r")) != (FILE *)NULL )
        { 
	  while( (n = getc(fd)) != EOF )  
	  {
            putc( n, ft->fd );
            putc( n, stdout );
          }
        }
        fclose( ft->fd );
      }
    }
    else if( f->transferType == TRANSFER_ASCII )
    {
      if( (ft->fd = fopen( ft->recvDir, "a" )) == (FILE *)NULL )
      {
        write( f->cmdSocketId, "ABOR\r\n", 6 );
        sprintf( buffer, "***ERROR: file descriptor for %s is NULL", ft->recvDir );
        popupErrorDialog( buffer );
        close( ft->dataSocketId );
        removeIthFtpDataConn( &f->ftpDataConn, i );
        return;
      }
      else
      {
        while( (n = read( *source, buffer, nb-1 )) > 0 )
        {
          buffer[n] = '\0';
	  fprintf( ft->fd, "%s", buffer );
        }
        fclose( ft->fd );
      }
    }
  }
  else 
  {
    if( (n = read( *source, buffer, nb-1 )) > 0 )
    {
      buffer[n] = '\0';
      ft->results = (char *)XtRealloc( (char *)ft->results,
	strlen(ft->results)+strlen(buffer)+2 );
      strcat( ft->results, buffer );
    }
    else
    {
      if( !strncmp( ft->cmd, "LIST", 4 ) )
      {
        parseLIST( ft->results, f, ft->index );

        if( f->cwdPtr )
        {
          char *strptr, buff[100];
          int j;
	  struct DirInfo *di;

          if( (strptr = strchr( f->cwdPtr, '/' )) )
    	  {
            strncpy( buff, f->cwdPtr, strptr-f->cwdPtr );
            buff[strptr-f->cwdPtr] = '\0';
            f->cwdPtr = strptr;
            if( f->cwdPtr && f->cwdPtr[0] == '/' )  f->cwdPtr++;
	  }
	  else
	  {
	    strcpy( buff, f->cwdPtr );
	    f->cwdPtr = strptr;
	  }

          di = getIthDir( f->dirInfo, f->cwdIndex );
          f->cwdIndex++;
          for( j = 0; di->nFiles; j++ )
          {
            struct FileList *flj = getIthFile( di->fileList, j );
            if( !strcmp( buff, flj->name ) )
            {
              XmListSelectPos( di->list, j+1, True );
              XmListSetPos( di->list, j+1 );
	      XFlush( XtDisplay(di->list) );
	      XSync( XtDisplay(di->list), False );
              break;
            }
          }
        }
        else
          printf("data results: %s\n", ft->results );
      }

      XtRemoveInput( *id );
      close( ft->acceptSocketId );
      fclose( ft->fd );
      XtFree( ft->results );
      XtFree( ft->cmd );
      removeIthFtpDataConn( &f->ftpDataConn, i );
    }
  }
}

void readFtpMessages( xtp, source, id )
  XtPointer xtp;
  int* source;
  XtInputId* id;
{
  char buffer[1056], buff[100];
  int n, nb = sizeof( buffer );
  struct FileManager *f = (struct FileManager *)xtp;

  memset( (void *)buffer, '\0', nb );

  if( (n = read( *source, buffer, nb-1 )) > 0 ) 
  {
    char ptr[1056], *strPtr;

    buffer[n] = '\0';
    strPtr = buffer;
    printf( "readFtpMessages() received -> %s\n", strPtr );

    while( sscanf( strPtr, "%[^\n]\n", ptr ) == 1 )
    {
      if( ptr[0] == '5' )
      {
        char errorMsg[1056];
        sprintf( errorMsg, "***FTP ERROR: %s", ptr );
        popupErrorDialog( errorMsg );
      }
      else if( !strncmp( ptr, "213", 3 ) )
      {
      }
      else if( !strncmp( ptr, "220", 3 ) )
      {
        sprintf( buff, "USER %s", f->user );
        write( f->cmdSocketId, buff, strlen(buff) );
        write( f->cmdSocketId, "\r\n", 2 );
      }
      else if( !strncmp( ptr, "331", 3 ) )
      {
        sprintf( buff, "PASS %s", f->pass );
        write( f->cmdSocketId, buff, strlen(buff) );
        write( f->cmdSocketId, "\r\n", 2 );
      }
      else if( !strncmp( ptr, "230", 3 ) )
      {
        struct stat st;
        addListObj( f );
        readDir( f, 0, "/", "/", &st, -1, 1, "dddddddddd" );
        ftpSendCommand( f, "PWD", 0, "" );
      }
      else if( !strncmp( ptr, "221", 3 ) )
      {
        deleteFileManager( f->mainForm );
      }
      else if( !strncmp( ptr, "425", 3 ) )
      {
        sprintf( buff, "Cannot open data connection to %s", f->host );
        popupErrorDialog( buff );
      }
      else if( strstr( ptr, "is current directory" ) )
      {
        char *strptr, buff[100];
        int j;
        struct DirInfo *di;

        f->cwd = (char *)XtMalloc( strlen(ptr)+1 );
        strcpy( f->cwd, strchr(ptr,'\"')+1 );
        strptr = strchr( f->cwd, '\"' );
        strptr[0] = '\0';
        f->cwdPtr = f->cwd + 1;

        strptr = strchr( f->cwdPtr, '/' );
        strncpy( buff, f->cwdPtr, strptr-f->cwdPtr );
        buff[strptr-f->cwdPtr] = '\0';
        f->cwdPtr = strptr;
        if( f->cwdPtr && f->cwdPtr[0] == '/' )  f->cwdPtr++;
        f->cwdIndex = 1;

        di = getIthDir( f->dirInfo, 0 );
        for( j = 0; di->nFiles; j++ )
        {
  	  struct FileList *flj = getIthFile( di->fileList, j );
          if( !strcmp( buff, flj->name ) )
          {
            XmListSelectPos( di->list, j+1, True );
            XmListSetPos( di->list, j+1 );
            break;
          }
        }
      }
      else
        printf( "message: %s\n", ptr );

      if( strlen(ptr)+1 >= strlen(strPtr) )
	break;
      else
	strPtr += strlen( ptr ) + 1;
    }
  }
  else 
  {
    XtRemoveInput( *id );
    close( f->cmdSocketId );
  }
}

void parseLIST( results, f, i )
  char *results;
  struct FileManager *f;
  int i;
{
  XmString xms;
  char *ptr = results, buff[1056];
  struct DirInfo *fd = getIthDir( f->dirInfo, i );
  struct FileList *fl;

  if( fd->list )  XmListDeleteAllItems( fd->list );

  while( sscanf( ptr, "%[^\r]", buff ) )
  {
    char perm[15], owner[50], group[50], month[5], yr_time[7], name[100];
    int lnks, size, day;

    if( !strcmp( buff, "" ) )  break;

    if( sscanf( buff, "%s %d %s %s %d %s %d %s %s", perm, &lnks,
	owner, group, &size, month, &day, yr_time, name ) == 9 )
    {
      char *tmpPtr;

      fl = appendFile( &fd->fileList );
      strcpy( fl->perm, perm );
      fd->nFiles++;

      if( (tmpPtr = strrchr( name, '/' )) )
      {
        strcpy( fl->name, tmpPtr+1 );
        strcpy( name, fl->name );
      }
      else
        strcpy( fl->name, name );

      if( perm[0] == 'l' )
	strcat( name, "@" );
      else if( perm[0] == 'd' )
        strcat( name, "/" );
      else if( perm[3] == 'x' || perm[6] == 'x' || perm[9] == 'x' )
	strcat( name, "*" );

      xms = XmStringCreateSimple( name );
      XmListAddItemUnselected( fd->list, xms, 0 );
      XmStringFree( xms );
    }

    ptr += strlen( buff );
    if( ptr )  
    {
	ptr += 1;
        if( ptr )  ptr += 1;
    }

    memset( (void *)buff, '\0', sizeof(buff) );
  }
}
